from __future__ import annotations

from dataclasses import dataclass, replace
from typing import Callable, Iterable, List, Literal, Optional, Tuple, TypeVar, Union

import pharmpy.tools.iivsearch.algorithms
from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.internals.set.subsets import non_empty_proper_subsets, non_empty_subsets
from pharmpy.model import Assignment, Model, RandomVariables
from pharmpy.modeling import add_iov, get_omegas, get_pk_parameters, remove_iiv, remove_iov
from pharmpy.modeling.parameter_variability import ADD_IOV_DISTRIBUTION
from pharmpy.tools import summarize_modelfit_results
from pharmpy.tools.common import (
    RANK_TYPES,
    ToolResults,
    create_results,
    summarize_tool,
    update_initial_estimates,
)
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder, call_workflow
from pharmpy.workflows.results import ModelfitResults

NAME_WF = 'iovsearch'

T = TypeVar('T')


def create_workflow(
    column: str = 'OCC',
    list_of_parameters: Optional[List[str]] = None,
    rank_type: Literal[tuple(RANK_TYPES)] = 'mbic',
    cutoff: Optional[Union[float, int]] = None,
    distribution: Literal[tuple(ADD_IOV_DISTRIBUTION)] = 'same-as-iiv',
    results: Optional[ModelfitResults] = None,
    model: Optional[Model] = None,
    strictness: Optional[str] = "minimization_successful or (rounding_errors and sigdigs>=0.1)",
):
    """Run IOVsearch tool. For more details, see :ref:`iovsearch`.

    Parameters
    ----------
    column : str
        Name of column in dataset to use as occasion column (default is 'OCC')
    list_of_parameters : None or list
        List of parameters to test IOV on, if none all parameters with IIV will be tested (default)
    rank_type : {'ofv', 'lrt', 'aic', 'bic', 'mbic'}
        Which ranking type should be used. Default is mBIC.
    cutoff : None or float
        Cutoff for which value of the ranking type that is considered significant. Default
        is None (all models will be ranked)
    distribution : {'disjoint', 'joint', 'explicit', 'same-as-iiv'}
        Which distribution added IOVs should have (default is same-as-iiv)
    results : ModelfitResults
        Results for model
    model : Model
        Pharmpy model
    strictness : str or None
        Strictness criteria

    Returns
    -------
    IOVSearchResults
        IOVSearch tool result object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model
    >>> from pharmpy.tools import run_iovsearch, load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> run_iovsearch('OCC', results=results, model=model)      # doctest: +SKIP
    """

    wb = WorkflowBuilder(name=NAME_WF)

    init_task = init(model, results)
    wb.add_task(init_task)

    bic_type = 'random'
    search_task = Task(
        'search',
        task_brute_force_search,
        column,
        list_of_parameters,
        rank_type,
        cutoff,
        bic_type,
        distribution,
    )

    wb.add_task(search_task, predecessors=init_task)
    search_output = wb.output_tasks

    results_task = Task(
        'results',
        task_results,
        rank_type,
        cutoff,
        bic_type,
        strictness,
    )

    wb.add_task(results_task, predecessors=search_output)

    return Workflow(wb)


def init(model, modelfit_results):
    return (
        Task('init', _model_entry, model)
        if model is None
        else Task('init', _model_entry, modelfit_results, model)
    )


def _model_entry(modelfit_results, model):
    return ModelEntry.create(model, modelfit_results=modelfit_results)


def task_brute_force_search(
    context,
    occ: str,
    list_of_parameters: Union[None, list],
    rank_type: str,
    cutoff: Union[None, float],
    bic_type: Union[None, str],
    distribution: str,
    input_model_entry: ModelEntry,
):
    input_model, input_res = input_model_entry.model, input_model_entry.modelfit_results
    # NOTE: Default is to try all IIV ETAs.
    if list_of_parameters is None:
        iiv = _get_nonfixed_iivs(input_model)
        iiv_before_odes = iiv.free_symbols.intersection(
            input_model.statements.before_odes.free_symbols
        )
        list_of_parameters = [iiv.name for iiv in iiv_before_odes]

    current_step = 0
    step_mapping = {current_step: [input_model.name]}

    # NOTE: Check that model has at least one IIV.
    if not list_of_parameters:
        return step_mapping, [input_model_entry]

    # NOTE: Add IOVs on given parameters or all parameters with IIVs.
    name = 'iovsearch_run1'
    model_with_iov = input_model.replace(name=name)
    model_with_iov = update_initial_estimates(model_with_iov, input_res)
    # TODO: Should we exclude already present IOVs?
    model_with_iov = add_iov(model_with_iov, occ, list_of_parameters, distribution=distribution)
    model_with_iov = model_with_iov.replace(description=_create_description(model_with_iov))
    # NOTE: Fit the new model.
    model_with_iov_entry = ModelEntry.create(model_with_iov, parent=input_model)
    wf = create_fit_workflow(modelentries=[model_with_iov_entry])
    model_with_iov_entry = call_workflow(wf, f'{NAME_WF}-fit-with-matching-IOVs', context)

    # NOTE: Remove IOVs. Test all subsets (~2^n).
    # TODO: Should we exclude already present IOVs?
    iov = model_with_iov.random_variables.iov
    # NOTE: We only need to remove the IOV ETA corresponding to the first
    # category in order to remove all IOV ETAs of the other categories
    all_iov_parameters = list(filter(lambda name: name.endswith('_1'), iov.names))
    no_of_models = 1
    wf = wf_etas_removal(
        remove_iov,
        model_with_iov_entry,
        non_empty_proper_subsets(all_iov_parameters),
        no_of_models + 1,
    )
    iov_candidate_entries = call_workflow(wf, f'{NAME_WF}-fit-with-removed-IOVs', context)

    # NOTE: Keep the best candidate.
    best_model_entry_so_far = best_model(
        input_model_entry,
        [model_with_iov_entry, *iov_candidate_entries],
        rank_type=rank_type,
        cutoff=cutoff,
        bic_type=bic_type,
    )

    current_step += 1
    step_mapping[current_step] = [model_with_iov.name] + [
        model_entry.model.name for model_entry in iov_candidate_entries
    ]

    # NOTE: If no improvement with respect to input model, STOP.
    if best_model_entry_so_far.model is input_model:
        return step_mapping, [input_model_entry, model_with_iov_entry, *iov_candidate_entries]

    # NOTE: Remove IIV with corresponding IOVs. Test all subsets (~2^n).
    iiv_parameters_with_associated_iov = list(
        map(
            lambda s: s.name,
            _get_iiv_etas_with_corresponding_iov(best_model_entry_so_far.model),
        )
    )
    # TODO: Should we exclude already present IOVs?
    no_of_models = len(iov_candidate_entries) + 1
    wf = wf_etas_removal(
        remove_iiv,
        best_model_entry_so_far,
        non_empty_subsets(iiv_parameters_with_associated_iov),
        no_of_models + 1,
    )
    iiv_candidate_entries = call_workflow(wf, f'{NAME_WF}-fit-with-removed-IIVs', context)
    current_step += 1
    step_mapping[current_step] = [model_entry.model.name for model_entry in iiv_candidate_entries]

    return step_mapping, [
        input_model_entry,
        model_with_iov_entry,
        *iov_candidate_entries,
        *iiv_candidate_entries,
    ]


def _create_description(model):
    iiv_desc = pharmpy.tools.iivsearch.algorithms.create_description(model)
    iov_desc = pharmpy.tools.iivsearch.algorithms.create_description(model, iov=True)
    return f'IIV({iiv_desc});IOV({iov_desc})'


def task_remove_etas_subset(
    remove: Callable[[Model, List[str]], None], model_entry: ModelEntry, subset: List[str], n: int
):
    parent_model, parent_res = model_entry.model, model_entry.modelfit_results
    model_with_some_etas_removed = parent_model.replace(name=f'iovsearch_run{n}')
    model_with_some_etas_removed = update_initial_estimates(
        model_with_some_etas_removed, parent_res
    )
    model_with_some_etas_removed = remove(model_with_some_etas_removed, subset)
    model_with_some_etas_removed = model_with_some_etas_removed.replace(
        description=_create_description(model_with_some_etas_removed)
    )
    return ModelEntry.create(model_with_some_etas_removed, parent=parent_model)


def wf_etas_removal(
    remove: Callable[[Model, List[str]], None],
    model_entry: ModelEntry,
    etas_subsets: Iterable[Tuple[str]],
    i: int,
):
    wb = WorkflowBuilder()
    j = i
    for subset_of_iiv_parameters in etas_subsets:
        task = Task(
            repr(subset_of_iiv_parameters),
            task_remove_etas_subset,
            remove,
            model_entry,
            list(subset_of_iiv_parameters),
            j,
        )
        wb.add_task(task)
        j += 1

    n = j - i
    wf_fit = create_fit_workflow(n=n)
    wb.insert_workflow(wf_fit)

    task_gather = Task('gather', lambda *model_entries: model_entries)
    wb.add_task(task_gather, predecessors=wb.output_tasks)
    return Workflow(wb)


def best_model(
    base_entry: ModelEntry,
    model_entries: List[ModelEntry],
    rank_type: str,
    cutoff: Union[None, float],
    bic_type: Union[None, str],
):
    candidate_entries = [base_entry, *model_entries]
    df = summarize_tool(
        model_entries, base_entry, rank_type=rank_type, cutoff=cutoff, bic_type=bic_type
    )
    best_model_name = df['rank'].idxmin()

    try:
        return [
            model_entry
            for model_entry in candidate_entries
            if model_entry.model.name == best_model_name
        ][0]
    except IndexError:
        return base_entry


def task_results(rank_type, cutoff, bic_type, strictness, step_mapping_and_model_entries):
    step_mapping, (base_model_entry, *res_model_entries) = step_mapping_and_model_entries

    model_dict = {
        model_entry.model.name: model_entry
        for model_entry in [base_model_entry] + res_model_entries
    }
    sum_mod, sum_tool = [], []
    for step, model_names in step_mapping.items():
        candidate_entries = [
            model_entry
            for model_name, model_entry in model_dict.items()
            if model_name in model_names
        ]
        sum_mod_step = summarize_modelfit_results(
            [model_entry.modelfit_results for model_entry in candidate_entries]
        )
        sum_mod.append(sum_mod_step)
        if step >= 1:
            ref_model_entry = model_dict[candidate_entries[0].parent.name]
            sum_tool_step = summarize_tool(
                candidate_entries, ref_model_entry, rank_type, cutoff, bic_type
            )
            sum_tool.append(sum_tool_step)

    keys = list(range(1, len(step_mapping)))

    res = create_results(
        IOVSearchResults,
        base_model_entry,
        base_model_entry,
        res_model_entries,
        rank_type,
        cutoff,
        bic_type=bic_type,
        summary_models=pd.concat(sum_mod, keys=[0] + keys, names=['step']),
        strictness=strictness,
    )

    # NOTE: This overwrites the default summary_tool field
    res = replace(res, summary_tool=pd.concat(sum_tool, keys=keys, names=['step']))

    return res


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(
    model,
    column,
    list_of_parameters,
    rank_type,
    distribution,
    strictness,
):
    if model is not None:
        if column not in model.datainfo.names:
            raise ValueError(
                f'Invalid `column`: got `{column}`,'
                f' must be one of {sorted(model.datainfo.names)}.'
            )

        if list_of_parameters is not None:
            allowed_parameters = set(get_pk_parameters(model)).union(
                str(statement.symbol) for statement in model.statements.before_odes
            )
            if not set(list_of_parameters).issubset(allowed_parameters):
                raise ValueError(
                    f'Invalid `list_of_parameters`: got `{list_of_parameters}`,'
                    f' must be NULL/None or a subset of {sorted(allowed_parameters)}.'
                )
    if strictness is not None and "rse" in strictness.lower():
        if model.estimation_steps[-1].parameter_uncertainty_method is None:
            raise ValueError(
                'parameter_uncertainty_method not set for model, cannot calculate relative standard errors.'
            )


@dataclass(frozen=True)
class IOVSearchResults(ToolResults):
    pass


def _get_iov_piecewise_assignment_symbols(model: Model):
    iovs = set(sympy.Symbol(rv) for rv in model.random_variables.iov.names)
    for statement in model.statements:
        if isinstance(statement, Assignment) and isinstance(statement.expression, sympy.Piecewise):
            try:
                expression_symbols = [p[0] for p in statement.expression.as_expr_set_pairs()]
            except (ValueError, NotImplementedError):
                pass  # NOTE: These exceptions are raised by complex Piecewise
                # statements that can be present in user code.
            else:
                if all(s in iovs for s in expression_symbols):
                    yield statement.symbol


def _get_iiv_etas_with_corresponding_iov(model: Model):
    iovs = set(_get_iov_piecewise_assignment_symbols(model))
    iiv = _get_nonfixed_iivs(model)

    for statement in model.statements:
        if isinstance(statement, Assignment) and isinstance(statement.expression, sympy.Add):
            for symbol in statement.expression.free_symbols:
                if symbol in iovs:
                    rest = statement.expression - symbol
                    if isinstance(rest, sympy.Symbol) and rest in iiv:
                        yield rest
                    break


def _get_nonfixed_iivs(model):
    fixed_omegas = get_omegas(model).fixed.names
    iivs = model.random_variables.iiv
    nonfixed_iivs = [
        iiv for iiv in iivs if str(list(iiv.variance.free_symbols)[0]) not in fixed_omegas
    ]
    return RandomVariables.create(nonfixed_iivs)



import pytest

from pharmpy.modeling import add_covariate_effect, get_covariates, remove_covariate_effect
from pharmpy.tools.covsearch.tool import (
    create_workflow,
    filter_search_space_and_model,
    validate_input,
)
from pharmpy.workflows import Workflow

MINIMAL_INVALID_MFL_STRING = ''
MINIMAL_VALID_MFL_STRING = 'LET(x, 0)'
LARGE_VALID_MFL_STRING = 'COVARIATE?(@IIV, @CONTINUOUS, *);COVARIATE?(@IIV, @CATEGORICAL, CAT)'


def test_create_workflow():
    assert isinstance(create_workflow(MINIMAL_VALID_MFL_STRING), Workflow)


def test_create_workflow_with_model(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'ruvsearch' / 'mox3.mod')
    assert isinstance(create_workflow(MINIMAL_VALID_MFL_STRING, model=model), Workflow)


def test_validate_input():
    validate_input(MINIMAL_VALID_MFL_STRING)


@pytest.mark.parametrize(
    ('model_path',), ((('nonmem', 'pheno.mod'),), (('nonmem', 'ruvsearch', 'mox3.mod'),))
)
def test_validate_input_with_model(load_model_for_test, testdata, model_path):
    model = load_model_for_test(testdata.joinpath(*model_path))
    validate_input(LARGE_VALID_MFL_STRING, model=model)


@pytest.mark.parametrize(
    (
        'model_path',
        'arguments',
        'exception',
        'match',
    ),
    [
        (
            None,
            dict(p_forward='x'),
            TypeError,
            'Invalid `p_forward`',
        ),
        (
            None,
            dict(p_forward=1.05),
            ValueError,
            'Invalid `p_forward`',
        ),
        (
            None,
            dict(p_backward=[]),
            TypeError,
            'Invalid `p_backward`',
        ),
        (
            None,
            dict(p_backward=1.01),
            ValueError,
            'Invalid `p_backward`',
        ),
        (
            None,
            dict(max_steps=1.2),
            TypeError,
            'Invalid `max_steps`',
        ),
        (None, dict(algorithm=()), ValueError, 'Invalid `algorithm`'),
        (
            None,
            dict(algorithm='scm-backward'),
            ValueError,
            'Invalid `algorithm`',
        ),
        (('nonmem', 'pheno.mod'), dict(effects=1), TypeError, 'Invalid `effects`'),
        (
            ('nonmem', 'pheno.mod'),
            dict(effects=MINIMAL_INVALID_MFL_STRING),
            ValueError,
            'Invalid `effects`',
        ),
        (
            ('nonmem', 'pheno.mod'),
            dict(effects='LAGTIME()'),
            ValueError,
            'Invalid `effects`',
        ),
        (
            ('nonmem', 'pheno.mod'),
            dict(effects='COVARIATE([CL, VC], WGT, EXP)'),
            ValueError,
            'Invalid `effects` because of invalid parameter',
        ),
        (
            ('nonmem', 'pheno.mod'),
            dict(effects='COVARIATE([CL, V], SEX, EXP)'),
            ValueError,
            'Invalid `effects` because of invalid covariate',
        ),
        (
            ('nonmem', 'pheno.mod'),
            dict(effects='COVARIATE([CL, V], WGT, [EXP, ABC])'),
            ValueError,
            'Invalid `effects` because of invalid effect function',
        ),
        (
            ('nonmem', 'pheno.mod'),
            dict(
                effects=(
                    ('CL', 'WGT', 'exp', '*'),
                    ('V', 'WGT', 'exp', '-'),
                )
            ),
            ValueError,
            'Invalid `effects`',
        ),
        (
            None,
            dict(model=1),
            TypeError,
            'Invalid `model`',
        ),
    ],
)
def test_validate_input_raises(
    load_model_for_test,
    testdata,
    model_path,
    arguments,
    exception,
    match,
):
    model = load_model_for_test(testdata.joinpath(*model_path)) if model_path else None

    harmless_arguments = dict(
        effects=MINIMAL_VALID_MFL_STRING,
    )

    kwargs = {**harmless_arguments, 'model': model, **arguments}

    with pytest.raises(exception, match=match):
        validate_input(**kwargs)


def test_covariate_filtering(load_model_for_test, testdata):
    effects = 'COVARIATE(@IIV, @CONTINUOUS, lin);COVARIATE?(@IIV, @CATEGORICAL, CAT)'
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    orig_cov = get_covariates(model)
    assert len(orig_cov) == 3

    eff, filtered_model = filter_search_space_and_model(effects, model)
    assert len(eff) == 2
    expected_cov_eff = set((('CL', 'APGR', 'cat', '*'), ('V', 'APGR', 'cat', '*')))
    assert set(eff.keys()) == expected_cov_eff
    assert len(get_covariates(filtered_model)) == 2

    for cov_effect in get_covariates(model):
        model = remove_covariate_effect(model, cov_effect[0], cov_effect[1].name)

    model = add_covariate_effect(model, 'CL', 'WGT', 'pow', '*')
    assert len(get_covariates(model)) == 1
    effects = 'COVARIATE([CL, V],WGT,pow,*)'
    eff, filtered_model = filter_search_space_and_model(effects, model)
    assert len(get_covariates(filtered_model)) == 2
    assert len(eff) == 0



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

from pharmpy.deps import altair as alt
from pharmpy.deps import pandas as pd
from pharmpy.workflows import Results


@dataclass(frozen=True)
class AMDResults(Results):
    final_model: Optional[str] = None
    summary_tool: Optional[Any] = None
    summary_models: Optional[Any] = None
    summary_individuals_count: Optional[pd.DataFrame] = None
    summary_errors: Optional[pd.DataFrame] = None
    final_model_parameter_estimates: Optional[pd.DataFrame] = None
    final_model_dv_vs_ipred_plot: Optional[alt.Chart] = None
    final_model_dv_vs_pred_plot: Optional[alt.Chart] = None
    final_model_cwres_vs_idv_plot: Optional[alt.Chart] = None
    final_model_abs_cwres_vs_ipred_plot: Optional[alt.Chart] = None



from .generic import AttrToken, AttrTree, GenericParser, NoSuchRuleException
from .ignored import with_ignored_tokens
from .missing import InsertMissing

__all__ = (
    'AttrTree',
    'AttrToken',
    'NoSuchRuleException',
    'GenericParser',
    'with_ignored_tokens',
    'InsertMissing',
)



import pytest

from pharmpy.internals.unicode import int_to_superscript, itemize_strings


@pytest.mark.parametrize(
    "x,expected",
    [
        (23, "²³"),
        (-1, "⁻¹"),
        (2, "²"),
    ],
)
def test_int_to_superscript(x, expected):
    assert int_to_superscript(x) == expected


@pytest.mark.parametrize(
    "x,expected",
    [
        (("cobolt",), "'cobolt'"),
        (("aa", "bb"), " one of 'aa' or 'bb'"),
        (("rock", "paper", "scissors"), " one of 'rock', 'paper' or 'scissors'"),
    ],
)
def test_itemize_strings(x, expected):
    assert itemize_strings(x) == expected



from collections.abc import Collection, Container, Iterable, Iterator, Sequence, Sized
from inspect import signature
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from ..unicode import itemize_strings


def with_runtime_arguments_type_check(fn):
    sig = signature(fn)
    parameters = sig.parameters.values()
    type_hints = None

    def _wrapped(*args, **kwargs):
        # NOTE: This delays loading annotations until first use
        nonlocal type_hints
        if type_hints is None:
            type_hints = get_type_hints(fn)

        for i, parameter in enumerate(parameters):
            if parameter.annotation is parameter.empty:
                # NOTE: Do not check anything if there is no annotation
                continue

            name = parameter.name
            value = _arg(parameter, i, args, kwargs)

            # NOTE: We cannot use parameter.annotation as this does not work
            # in combination with `from __future__ import annotations`.
            # See https://peps.python.org/pep-0563/#introducing-a-new-dictionary-for-the-string-literal-form-instead
            expected_types = _annotation_to_types(type_hints[name])
            if not any(map(lambda expected_type: _match(expected_type, value), expected_types)):
                if len(expected_types) == 1 and get_origin(expected_types[0]) == Literal:
                    allowed_values = list(get_args(expected_types[0]))
                    raise ValueError(
                        f"Invalid `{parameter.name}`: got '{value}', must be"
                        + f"{itemize_strings(sorted(allowed_values))}"
                    )
                else:
                    raise TypeError(
                        f'Invalid `{parameter.name}`: got `{value}` of type {_value_type(value)},'
                        + (
                            f' expected {expected_types[0]}'
                            if len(expected_types) == 1
                            else f' expected one of {expected_types}.'
                        )
                    )

        return fn(*args, **kwargs)

    return _wrapped


def _type(t):
    return type(None) if t is None else t


def _annotation_to_types(annotation):
    types = annotation if isinstance(annotation, list) else [annotation]
    return [_type(t) for t in types]


def _value_type(value):
    return Type[value] if isinstance(value, type) else type(value)


def _kwargs(parameter, kwargs):
    try:
        return kwargs[parameter.name]
    except KeyError:
        default = parameter.default
        if default == parameter.empty:
            raise KeyError
        return default


def _arg(parameter, i, args, kwargs):
    if parameter.kind == parameter.POSITIONAL_ONLY:
        return args[i]
    elif parameter.kind == parameter.POSITIONAL_OR_KEYWORD:
        if i < len(args):
            return args[i]
        return _kwargs(parameter, kwargs)
    elif parameter.kind == parameter.VAR_POSITIONAL:
        raise NotImplementedError(parameter.kind)
    elif parameter.kind == parameter.KEYWORD_ONLY:
        return _kwargs(parameter, kwargs)
    else:
        assert parameter.kind == parameter.VAR_KEYWORD
        raise NotImplementedError(parameter.kind)


def _match_sequence_items(args, value):
    if args:
        assert len(args) == 1
        t = args[0]
        return all(map(lambda v: _match(t, v), value))
    else:
        return True


def _match(typing, value):
    origin = get_origin(typing)

    if origin is None:
        if typing is Any or typing is Optional:
            return True
        return isinstance(value, typing)

    if origin is Literal:
        # NOTE: Empty literals return False
        return any(map(lambda t: value == t, get_args(typing)))

    if origin is list or origin is List:
        return isinstance(value, list) and _match_sequence_items(get_args(typing), value)

    if origin is Sequence:
        return isinstance(value, Sequence) and _match_sequence_items(get_args(typing), value)

    if origin is tuple or origin is Tuple:
        if not isinstance(value, tuple):
            return False

        args = get_args(typing)
        n = len(args)
        if n == 2 and args[1] is Ellipsis:
            return _match_sequence_items((args[0],), value)

        else:
            return len(value) == n and all(map(_match, args, value))

    if origin is Union:
        # NOTE: Empty unions return False
        return any(map(lambda t: _match(t, value), get_args(typing)))

    if origin is Optional:
        args = get_args(typing)
        if args:
            assert len(args) == 1
            t = args[0]
            return value is None or _match(t, value)
        else:
            return True

    if origin is Sized:
        try:
            return isinstance(len(value), int)
        except TypeError:
            return False

    if origin is Container:
        # NOTE: Cannot check value type because we do not know any candidate key
        return hasattr(value, '__contains__')

    if origin is Iterator:
        # NOTE: Cannot check value type because we do not know any candidate key
        return hasattr(value, '__next__') and hasattr(value, '__iter__')

    if origin is Iterable:
        # NOTE: Cannot check value type because of risk of side-effect
        try:
            iter(value)
            return True
        except TypeError:
            return False

    if origin is Collection:
        t = get_args(typing)
        return (
            _match(Sized, value)
            and _match(
                Container[t] if t else Container,  # pyright: ignore [reportGeneralTypeIssues]
                value,
            )
            and _match(
                Iterable[t] if t else Iterable,  # pyright: ignore [reportGeneralTypeIssues]
                value,
            )
        )

    if origin is dict or origin is Dict:
        args = get_args(typing)
        if value is not None:
            return _match_sequence_items((args[0],), value.keys()) & _match_sequence_items(
                (args[1],), value.values()
            )
        else:
            return True

    raise NotImplementedError(origin)


def check_list(name, value, valid_values):
    """Check if value is in valid_values else raise"""
    if value not in valid_values:
        raise ValueError(
            f"Invalid `{name}`: got '{value}', must be{itemize_strings(sorted(valid_values))}"
        )



import os
import re
import shutil
from operator import add, mul

import pytest
from sympy import Symbol as S

from pharmpy.deps import sympy
from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Assignment, NormalDistribution
from pharmpy.modeling import (
    add_iiv,
    add_individual_parameter,
    add_iov,
    add_peripheral_compartment,
    add_pk_iiv,
    create_joint_distribution,
    remove_iiv,
    remove_iov,
    set_first_order_absorption,
    set_transit_compartments,
    set_zero_order_elimination,
    split_joint_distribution,
    transform_etas_boxcox,
    transform_etas_john_draper,
    transform_etas_tdist,
    update_initial_individual_estimates,
)
from pharmpy.modeling.parameter_variability import (
    EtaAddition,
    EtaTransformation,
    _choose_cov_param_init,
)
from pharmpy.tools import read_modelfit_results


@pytest.mark.parametrize(
    'parameter, expression, operation, eta_name, buf_new, no_of_omega_recs',
    [
        ('S1', 'exp', '+', None, 'V=TVV*EXP(ETA(2))\nS1 = V + EXP(ETA_S1)', 2),
        ('S1', 'exp', '*', None, 'V=TVV*EXP(ETA(2))\nS1 = V*EXP(ETA_S1)', 2),
        ('V', 'exp', '+', None, 'V = TVV*EXP(ETA(2)) + EXP(ETA_V)\nS1=V', 2),
        ('S1', 'add', None, None, 'V=TVV*EXP(ETA(2))\nS1 = V + ETA_S1', 2),
        ('S1', 'prop', None, None, 'V=TVV*EXP(ETA(2))\nS1 = V*(ETA_S1 + 1)', 2),
        ('S1', 'log', None, None, 'V=TVV*EXP(ETA(2))\nS1 = V*EXP(ETA_S1)/(EXP(ETA_S1) + 1)', 2),
        ('S1', 'eta_new', '+', None, 'V=TVV*EXP(ETA(2))\nS1 = V + ETA_S1', 2),
        ('S1', 'eta_new**2', '+', None, 'V=TVV*EXP(ETA(2))\nS1 = V + ETA_S1**2', 2),
        ('S1', 'exp', '+', 'ETA(3)', 'V=TVV*EXP(ETA(2))\nS1 = V + EXP(ETA(3))', 2),
        (
            'S1',
            're_log',
            '+',
            None,
            'V=TVV*EXP(ETA(2))\nPHI_V = LOG(V/(1 - V))'
            '\nS1 = EXP(ETA_S1*PHI_V)/(EXP(ETA_S1*PHI_V) + 1)',
            2,
        ),
        (
            ['V', 'S1'],
            'exp',
            '+',
            None,
            'V = TVV*EXP(ETA(2)) + EXP(ETA_V)\nS1 = V + EXP(ETA_S1)',
            3,
        ),
        (
            ['V', 'S1'],
            'exp',
            '+',
            ['new_eta1', 'new_eta2'],
            'V = TVV*EXP(ETA(2)) + EXP(NEW_ETA1)\nS1 = V + EXP(NEW_ETA2)',
            3,
        ),
    ],
)
def test_add_iiv(
    load_model_for_test,
    pheno_path,
    parameter,
    expression,
    operation,
    eta_name,
    buf_new,
    no_of_omega_recs,
):
    model = load_model_for_test(pheno_path)

    model = add_iiv(
        model,
        list_of_parameters=parameter,
        expression=expression,
        operation=operation,
        eta_names=eta_name,
    )

    etas = model.random_variables.etas.names

    assert eta_name is None or set(eta_name).intersection(etas) or eta_name in etas

    rec_ref = (
        f'$PK\n'
        f'IF(AMT.GT.0) BTIME=TIME\n'
        f'TAD=TIME-BTIME\n'
        f'TVCL=THETA(1)*WGT\n'
        f'TVV=THETA(2)*WGT\n'
        f'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n'
        f'CL=TVCL*EXP(ETA(1))\n'
        f'{buf_new}\n\n'
    )

    assert str(model.internals.control_stream.get_pred_pk_record()) == rec_ref

    omega_rec = model.internals.control_stream.get_records('OMEGA')

    assert len(omega_rec) == no_of_omega_recs
    assert '$OMEGA  0.09 ; IIV_' in str(omega_rec[-1])


def test_add_iiv_missing_param(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    with pytest.raises(ValueError):
        add_iiv(model, 'non_existing_param', 'add')


@pytest.mark.parametrize(
    'path, occ, etas, eta_names, pk_start_ref, pk_end_ref, omega_ref, distribution',
    [
        (
            'nonmem/pheno_real.mod',
            'FA1',
            ['ETA_1'],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'disjoint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            'ETA_1',
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'disjoint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            'ETA_1',
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'joint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            ['CL', 'ETA_1'],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'joint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            ['ETA_1', 'CL'],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'joint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            [['CL', 'ETA_1']],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'explicit',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            [['ETA_1', 'CL']],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'explicit',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            None,
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'IOV_2 = 0\n'
            'IF (FA1.EQ.0) IOV_2 = ETA_IOV_2_1\n'
            'IF (FA1.EQ.1) IOV_2 = ETA_IOV_2_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n'
            'ETAI2 = IOV_2 + ETA(2)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V = TVV*EXP(ETAI2)\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n'
            '0.00309626 ; OMEGA_IOV_1\n'
            '$OMEGA  BLOCK(1) SAME\n'
            '$OMEGA  BLOCK(1)\n'
            '0.0031128 ; OMEGA_IOV_2\n'
            '$OMEGA  BLOCK(1) SAME\n',
            'disjoint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            None,
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'IOV_2 = 0\n'
            'IF (FA1.EQ.0) IOV_2 = ETA_IOV_2_1\n'
            'IF (FA1.EQ.1) IOV_2 = ETA_IOV_2_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n'
            'ETAI2 = IOV_2 + ETA(2)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V = TVV*EXP(ETAI2)\n' 'S1=V\n',
            '$OMEGA BLOCK(2)\n'
            '0.00309626\t; OMEGA_IOV_1\n'
            '0.001\t; OMEGA_IOV_1_2\n'
            '0.0031128\t; OMEGA_IOV_2\n'
            '$OMEGA BLOCK(2) SAME\n',
            'joint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            None,
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'IOV_2 = 0\n'
            'IF (FA1.EQ.0) IOV_2 = ETA_IOV_2_1\n'
            'IF (FA1.EQ.1) IOV_2 = ETA_IOV_2_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n'
            'ETAI2 = IOV_2 + ETA(2)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V = TVV*EXP(ETAI2)\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n'
            '0.00309626 ; OMEGA_IOV_1\n'
            '$OMEGA  BLOCK(1) SAME\n'
            '$OMEGA  BLOCK(1)\n'
            '0.0031128 ; OMEGA_IOV_2\n'
            '$OMEGA  BLOCK(1) SAME\n',
            'same-as-iiv',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            ['ETA_2'],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA(2)\n',
            'CL=TVCL*EXP(ETA(1))\n' 'V = TVV*EXP(ETAI1)\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.0031128 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'same-as-iiv',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            ['CL'],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'disjoint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            ['CL', 'ETA_2'],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'IOV_2 = 0\n'
            'IF (FA1.EQ.0) IOV_2 = ETA_IOV_2_1\n'
            'IF (FA1.EQ.1) IOV_2 = ETA_IOV_2_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n'
            'ETAI2 = IOV_2 + ETA(2)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V = TVV*EXP(ETAI2)\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n'
            '0.00309626 ; OMEGA_IOV_1\n'
            '$OMEGA  BLOCK(1) SAME\n'
            '$OMEGA  BLOCK(1)\n'
            '0.0031128 ; OMEGA_IOV_2\n'
            '$OMEGA  BLOCK(1) SAME\n',
            'disjoint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            ['CL', 'ETA_2'],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'IOV_2 = 0\n'
            'IF (FA1.EQ.0) IOV_2 = ETA_IOV_2_1\n'
            'IF (FA1.EQ.1) IOV_2 = ETA_IOV_2_2\n'
            'ETAI1 = IOV_1 + ETA(1)\n'
            'ETAI2 = IOV_2 + ETA(2)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V = TVV*EXP(ETAI2)\n' 'S1=V\n',
            '$OMEGA BLOCK(2)\n'
            '0.00309626\t; OMEGA_IOV_1\n'
            '0.001\t; OMEGA_IOV_1_2\n'
            '0.0031128\t; OMEGA_IOV_2\n'
            '$OMEGA BLOCK(2) SAME\n',
            'joint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            ['ETA_1'],
            ['ETA_3', 'ETA_4'],
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA(3)\n'
            'IF (FA1.EQ.1) IOV_1 = ETA(4)\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'disjoint',
        ),
        (
            'nonmem/pheno_real.mod',
            'FA1',
            ['ETA_1'],
            ['ETA_3', 'ETA_4'],
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA(3)\n'
            'IF (FA1.EQ.1) IOV_1 = ETA(4)\n'
            'ETAI1 = IOV_1 + ETA(1)\n',
            'CL = TVCL*EXP(ETAI1)\n' 'V=TVV*EXP(ETA(2))\n' 'S1=V\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'joint',
        ),
        (
            'nonmem/pheno_block.mod',
            'FA1',
            ['ETA_CL'],
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'ETAI1 = IOV_1 + ETA_CL\n',
            'CL = THETA(1)*EXP(ETAI1)\n'
            'V=THETA(2)*EXP(ETA_V)\n'
            'S1=V+ETA_S1\n'
            'MAT=THETA(3)*EXP(ETA_MAT)\n'
            'Q=THETA(4)*EXP(ETA_Q)\n',
            '$OMEGA  BLOCK(1)\n' '0.00309626 ; OMEGA_IOV_1\n' '$OMEGA  BLOCK(1) SAME\n',
            'disjoint',
        ),
        (
            'nonmem/pheno_block.mod',
            'FA1',
            None,
            None,
            'IOV_1 = 0\n'
            'IF (FA1.EQ.0) IOV_1 = ETA_IOV_1_1\n'
            'IF (FA1.EQ.1) IOV_1 = ETA_IOV_1_2\n'
            'IOV_2 = 0\n'
            'IF (FA1.EQ.0) IOV_2 = ETA_IOV_2_1\n'
            'IF (FA1.EQ.1) IOV_2 = ETA_IOV_2_2\n'
            'IOV_3 = 0\n'
            'IF (FA1.EQ.0) IOV_3 = ETA_IOV_3_1\n'
            'IF (FA1.EQ.1) IOV_3 = ETA_IOV_3_2\n'
            'IOV_4 = 0\n'
            'IF (FA1.EQ.0) IOV_4 = ETA_IOV_4_1\n'
            'IF (FA1.EQ.1) IOV_4 = ETA_IOV_4_2\n'
            'IOV_5 = 0\n'
            'IF (FA1.EQ.0) IOV_5 = ETA_IOV_5_1\n'
            'IF (FA1.EQ.1) IOV_5 = ETA_IOV_5_2\n'
            'ETAI1 = IOV_1 + ETA_CL\n'
            'ETAI2 = IOV_2 + ETA_V\n'
            'ETAI3 = IOV_3 + ETA_S1\n'
            'ETAI4 = IOV_4 + ETA_MAT\n'
            'ETAI5 = IOV_5 + ETA_Q\n',
            'CL = THETA(1)*EXP(ETAI1)\n'
            'V = THETA(2)*EXP(ETAI2)\n'
            'S1 = ETAI3 + V\n'
            'MAT = THETA(3)*EXP(ETAI4)\n'
            'Q = THETA(4)*EXP(ETAI5)\n',
            '$OMEGA  BLOCK(1)\n'
            '0.00309626 ; OMEGA_IOV_1\n'
            '$OMEGA  BLOCK(1) SAME\n'
            '$OMEGA  BLOCK(1)\n'
            '0.0031128 ; OMEGA_IOV_2\n'
            '$OMEGA  BLOCK(1) SAME\n'
            '$OMEGA  BLOCK(1)\n'
            '0.010000000000000002 ; OMEGA_IOV_3\n'
            '$OMEGA  BLOCK(1) SAME\n'
            '$OMEGA BLOCK(2)\n'
            '0.00309626\t; OMEGA_IOV_4\n'
            '5E-05\t; OMEGA_IOV_4_5\n'
            '0.0031128\t; OMEGA_IOV_5\n'
            '$OMEGA BLOCK(2) SAME\n',
            'same-as-iiv',
        ),
    ],
)
def test_add_iov(
    load_model_for_test,
    testdata,
    path,
    occ,
    etas,
    eta_names,
    pk_start_ref,
    pk_end_ref,
    omega_ref,
    distribution,
):
    model = load_model_for_test(testdata / path)
    model = add_iov(model, occ, etas, eta_names, distribution=distribution)

    model_etas = set(model.random_variables.etas.names)
    assert eta_names is None or model_etas.issuperset(eta_names)

    pk_rec = str(model.internals.control_stream.get_pred_pk_record())

    expected_pk_rec_start = f'$PK\n{pk_start_ref}'
    expected_pk_rec_end = f'{pk_end_ref}\n'

    assert pk_rec[: len(expected_pk_rec_start)] == expected_pk_rec_start
    assert pk_rec[-len(expected_pk_rec_end) :] == expected_pk_rec_end

    rec_omega = ''.join(str(rec) for rec in model.internals.control_stream.get_records('OMEGA'))

    assert rec_omega[-len(omega_ref) :] == omega_ref

    if eta_names:
        assert len(model.internals.control_stream.get_records('ABBREVIATED')) == 0
    else:
        assert len(model.internals.control_stream.get_records('ABBREVIATED')) > 0


def test_add_iov_compose(load_model_for_test, pheno_path):
    model1 = load_model_for_test(pheno_path)
    model1 = add_iov(model1, 'FA1', ['ETA_1', 'ETA_2'])

    model2 = load_model_for_test(pheno_path)
    model2 = add_iov(model2, 'FA1', 'ETA_1')
    model2 = add_iov(model2, 'FA1', 'ETA_2')

    assert set(model1.random_variables.etas.names) == set(model2.random_variables.etas.names)
    # FIXME: Find a better way to assert that models are equivalent
    assert sorted(str(model1.internals.control_stream.get_pred_pk_record()).split('\n')) == sorted(
        str(model2.internals.control_stream.get_pred_pk_record()).split('\n')
    )

    rec_omega_1 = list(str(rec) for rec in model1.internals.control_stream.get_records('OMEGA'))
    rec_omega_2 = list(str(rec) for rec in model2.internals.control_stream.get_records('OMEGA'))

    assert rec_omega_1 == rec_omega_2


def test_add_iov_only_one_level(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    df = model.dataset.copy()
    df['FA1'] = 1
    model = model.replace(dataset=df)

    with pytest.raises(ValueError, match='Only one value in FA1 column.'):
        add_iov(model, 'FA1', ['ETA_1'])


@pytest.mark.parametrize(
    'occ, params, new_eta_names, distribution, error, message',
    (
        (
            'FA1',
            ['ETA_1', 'CL'],
            None,
            'disjoint',
            ValueError,
            'ETA_1 was given twice.',
        ),
        (
            'FA1',
            ['CL', 'ETA_1'],
            None,
            'disjoint',
            ValueError,
            'ETA_1 was given twice.',
        ),
        (
            'FA1',
            [['ETA_1'], ['CL']],
            None,
            'explicit',
            ValueError,
            'ETA_1 was given twice.',
        ),
        (
            'FA1',
            [['CL'], ['ETA_1']],
            None,
            'explicit',
            ValueError,
            'ETA_1 was given twice.',
        ),
        (
            'FA1',
            ['ETA_1'],
            None,
            'abracadabra',
            ValueError,
            '"abracadabra" is not a valid value for distribution',
        ),
        (
            'FA1',
            ['ETA_1'],
            None,
            'explicit',
            ValueError,
            'distribution == "explicit" requires parameters to be given as lists of lists',
        ),
        (
            'FA1',
            [['ETA_2'], 'ETA_1'],
            None,
            'explicit',
            ValueError,
            'distribution == "explicit" requires parameters to be given as lists of lists',
        ),
        (
            'FA1',
            [['ETA_1']],
            None,
            'joint',
            ValueError,
            'distribution != "explicit" requires parameters to be given as lists of strings',
        ),
        (
            'FA1',
            [['ETA_1'], [2, 'ETA_2']],
            None,
            'explicit',
            ValueError,
            'not all parameters are string',
        ),
        (
            'FA1',
            [['ETA_1', 'ETA_2']],
            ['A', 'B', 'C', 'D', 'E'],
            'explicit',
            ValueError,
            'Number of given eta names is incorrect, need 4 names.',
        ),
    ),
)
def test_add_iov_raises(
    load_model_for_test, pheno_path, occ, params, new_eta_names, distribution, error, message
):
    model = load_model_for_test(pheno_path)
    with pytest.raises(error, match=re.escape(message)):
        add_iov(model, occ, params, eta_names=new_eta_names, distribution=distribution)


@pytest.mark.parametrize(
    'addition,expression',
    [
        (EtaAddition.exponential(add), S('CL') + sympy.exp(S('eta_new'))),
        (EtaAddition.exponential(mul), S('CL') * sympy.exp(S('eta_new'))),
    ],
)
def test_add_iiv_apply(addition, expression):
    addition.apply(original='CL', eta='eta_new')

    assert addition.template == expression


@pytest.mark.parametrize(
    'eta_iov_1,eta_iov_2',
    [
        ('ETA_IOV_1_1', 'ETA_IOV_2_1'),
        ('ETA_IOV_1_1', 'ETA_IOV_3_1'),
        ('ETA_IOV_2_1', 'ETA_IOV_3_1'),
        ('ETA_IOV_2_1', 'ETA_IOV_1_1'),
        ('ETA_IOV_3_1', 'ETA_IOV_1_1'),
        ('ETA_IOV_3_1', 'ETA_IOV_2_1'),
    ],
)
def test_add_iov_regression_code_record(load_model_for_test, testdata, eta_iov_1, eta_iov_2):
    """This is a regression test for NONMEM CodeRecord statements property round-trip
    serialization.
    See https://github.com/pharmpy/pharmpy/pull/771
    """

    model_no_iov = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox2.mod')
    model = add_iov(model_no_iov, occ="VISI")

    # Remove the first IOV, can reproduce the same issue
    model_r1 = remove_iov(model, to_remove=[eta_iov_1])

    # Remove the second IOV, can reproduce the same issue
    model_r1r2 = remove_iov(model_r1, to_remove=[eta_iov_2])

    # Remove the first and second IOV
    model_r12 = remove_iov(model, to_remove=[eta_iov_1, eta_iov_2])

    assert model_r12 == model_r1r2
    assert model_r12.model_code == model_r1r2.model_code

    model_r2r1 = remove_iov(model, to_remove=[eta_iov_2])
    model_r2r1 = remove_iov(model_r2r1, to_remove=[eta_iov_1])

    assert model_r1r2 == model_r2r1
    assert model_r1r2.model_code == model_r2r1.model_code


def test_add_pk_iiv_1(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    model = set_zero_order_elimination(model)
    model = add_pk_iiv(model)
    iivs = set(model.random_variables.iiv.names)
    assert iivs == {'ETA_1', 'ETA_2', 'ETA_KM'}
    model = add_peripheral_compartment(model)
    model = add_pk_iiv(model)
    iivs = set(model.random_variables.iiv.names)
    assert iivs == {'ETA_1', 'ETA_2', 'ETA_KM', 'ETA_VP1', 'ETA_QP1'}


def test_add_pk_iiv_2(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    model = set_zero_order_elimination(model)
    model = add_peripheral_compartment(model)
    model = add_pk_iiv(model)
    iivs = set(model.random_variables.iiv.names)
    assert iivs == {'ETA_1', 'ETA_2', 'ETA_KM', 'ETA_VP1', 'ETA_QP1'}


def test_add_pk_iiv_nested_params(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    model = set_transit_compartments(model, 3)
    model = add_pk_iiv(model)
    iivs = set(model.random_variables.iiv.names)
    assert iivs == {'ETA_1', 'ETA_2', 'ETA_MDT'}

    model = load_model_for_test(pheno_path)
    model = set_first_order_absorption(model)
    model = add_pk_iiv(model)
    iivs = set(model.random_variables.iiv.names)
    assert iivs == {'ETA_1', 'ETA_2', 'ETA_MAT'}

    model = load_model_for_test(pheno_path)
    model = set_transit_compartments(model, 3)
    model = add_pk_iiv(model, initial_estimate=0.01)
    assert model.parameters['IIV_MDT'].init == 0.01


@pytest.mark.parametrize(
    'etas, pk_ref, omega_ref',
    [
        (
            ['ETA_CL'],
            '$PK\n'
            'CL = THETA(1)\n'
            'V=THETA(2)*EXP(ETA_V)\n'
            'S1=V+ETA_S1\n'
            'MAT=THETA(3)*EXP(ETA_MAT)\n'
            'Q=THETA(4)*EXP(ETA_Q)\n\n',
            '$OMEGA 0.031128  ; IVV\n'
            '$OMEGA 0.1\n'
            '$OMEGA BLOCK(2)\n'
            '0.0309626\n'
            '0.0005 0.031128\n',
        ),
        (
            ['ETA_CL', 'ETA_V'],
            '$PK\n'
            'CL = THETA(1)\n'
            'V = THETA(2)\n'
            'S1=V+ETA_S1\n'
            'MAT=THETA(3)*EXP(ETA_MAT)\n'
            'Q=THETA(4)*EXP(ETA_Q)\n\n',
            '$OMEGA 0.1\n' '$OMEGA BLOCK(2)\n' '0.0309626\n' '0.0005 0.031128\n',
        ),
        (
            ['ETA_CL', 'ETA_MAT'],
            '$PK\n'
            'CL = THETA(1)\n'
            'V=THETA(2)*EXP(ETA_V)\n'
            'S1=V+ETA_S1\n'
            'MAT = THETA(3)\n'
            'Q=THETA(4)*EXP(ETA_Q)\n\n',
            '$OMEGA 0.031128  ; IVV\n' '$OMEGA 0.1\n' '$OMEGA  0.031128 ; OMEGA_5_5\n',
        ),
        (
            ['ETA_MAT', 'ETA_Q'],
            '$PK\n'
            'CL=THETA(1)*EXP(ETA_CL)\n'
            'V=THETA(2)*EXP(ETA_V)\n'
            'S1=V+ETA_S1\n'
            'MAT = THETA(3)\n'
            'Q = THETA(4)\n\n',
            '$OMEGA DIAGONAL(2)\n' '0.0309626  ; IVCL\n' '0.031128  ; IVV\n' '$OMEGA 0.1\n',
        ),
        (
            None,
            '$PK\n'
            'DUMMYETA = ETA(1)\n'
            'CL = THETA(1)\n'
            'V = THETA(2)\n'
            'S1 = V\n'
            'MAT = THETA(3)\n'
            'Q = THETA(4)\n\n',
            '$OMEGA  0 FIX ; DUMMYOMEGA\n',
        ),
        (
            ['CL'],
            '$PK\n'
            'CL = THETA(1)\n'
            'V=THETA(2)*EXP(ETA_V)\n'
            'S1=V+ETA_S1\n'
            'MAT=THETA(3)*EXP(ETA_MAT)\n'
            'Q=THETA(4)*EXP(ETA_Q)\n\n',
            '$OMEGA 0.031128  ; IVV\n'
            '$OMEGA 0.1\n'
            '$OMEGA BLOCK(2)\n'
            '0.0309626\n'
            '0.0005 0.031128\n',
        ),
        (
            'ETA_CL',
            '$PK\n'
            'CL = THETA(1)\n'
            'V=THETA(2)*EXP(ETA_V)\n'
            'S1=V+ETA_S1\n'
            'MAT=THETA(3)*EXP(ETA_MAT)\n'
            'Q=THETA(4)*EXP(ETA_Q)\n\n',
            '$OMEGA 0.031128  ; IVV\n'
            '$OMEGA 0.1\n'
            '$OMEGA BLOCK(2)\n'
            '0.0309626\n'
            '0.0005 0.031128\n',
        ),
    ],
)
def test_remove_iiv(load_model_for_test, testdata, etas, pk_ref, omega_ref):
    model = load_model_for_test(testdata / 'nonmem/pheno_block.mod')
    model = remove_iiv(model, etas)

    assert str(model.internals.control_stream.get_pred_pk_record()) == pk_ref

    rec_omega = ''.join(str(rec) for rec in model.internals.control_stream.get_records('OMEGA'))

    assert rec_omega == omega_ref


@pytest.mark.parametrize(
    'iiv_type, operation',
    [
        ('add', '*'),
        ('prop', '*'),
        ('log', '*'),
        ('exp', '*'),
        ('exp', '+'),
    ],
)
def test_remove_iiv2(load_model_for_test, testdata, iiv_type, operation):
    model = load_model_for_test(testdata / 'nonmem/pheno_real.mod')

    model = add_individual_parameter(model, 'A')
    without_iiv = model.statements.find_assignment('A')
    model = add_iiv(model, 'A', iiv_type, operation)
    model = remove_iiv(model, 'A')
    assert model.statements.find_assignment('A') == without_iiv

    model = add_iiv(model, "TVCL", "add")
    model = remove_iiv(model, 'TVCL')
    assert model.statements.find_assignment('TVCL') == Assignment(
        sympy.Symbol('TVCL'), sympy.Symbol('PTVCL') * sympy.Symbol('WGT')
    )

    # Test with two different ETAs
    model6 = add_iiv(model, 'CL', 'add')
    model6 = remove_iiv(model6, 'CL')
    assert model6.statements.find_assignment('CL') == Assignment(
        sympy.Symbol('CL'), sympy.Symbol('TVCL')
    )

    model7 = add_iiv(model, 'Y', 'exp')
    model7 = remove_iiv(model7, 'Y')
    assert model7.statements.find_assignment('Y') == model.statements.find_assignment('Y')

    model8 = add_iiv(model, 'Y', 'add')
    model8 = remove_iiv(model7, 'Y')
    assert model8.statements.find_assignment('Y') == model.statements.find_assignment('Y')

    # Check that exp(ETA + IOV) becomes exp(IOV) after removing iiv
    model = load_model_for_test(testdata / 'nonmem/models/fviii6.mod')
    model = remove_iiv(model, 'CL')
    assert model.statements.find_assignment('CL') == Assignment(
        sympy.Symbol('CL'), sympy.Symbol('TVCL') * sympy.exp(sympy.Symbol('IOVCL'))
    )


def test_remove_iov(create_model_for_test, load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem/pheno_block.mod')

    model_str = model.model_code
    model_with_iov = model_str.replace(
        '$OMEGA DIAGONAL(2)\n' '0.0309626  ; IVCL\n' '0.031128  ; IVV',
        '$OMEGA BLOCK(1)\n0.1\n$OMEGA BLOCK(1) SAME\n',
    )

    model = create_model_for_test(model_with_iov)

    model = remove_iov(model)

    assert (
        str(model.internals.control_stream.get_pred_pk_record()) == '$PK\n'
        'CL = THETA(1)\n'
        'V = THETA(2)\n'
        'S1=V+ETA_S1\n'
        'MAT=THETA(3)*EXP(ETA_MAT)\n'
        'Q=THETA(4)*EXP(ETA_Q)\n\n'
    )
    rec_omega = ''.join(str(rec) for rec in model.internals.control_stream.get_records('OMEGA'))

    assert rec_omega == '$OMEGA 0.1\n' '$OMEGA BLOCK(2)\n' '0.0309626\n' '0.0005 0.031128\n'


def test_remove_iov_no_iovs(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem/pheno_block.mod')

    with pytest.warns(UserWarning):
        remove_iov(model)


def test_remove_iov_github_issues_538_and_561_1(load_model_for_test, testdata):
    m = load_model_for_test(testdata / 'nonmem' / 'models' / 'fviii6.mod')

    m = remove_iov(m)

    assert not m.random_variables.iov


def test_remove_iov_github_issues_538_and_561_2(load_model_for_test, testdata):
    m = load_model_for_test(testdata / 'nonmem' / 'models' / 'fviii6.mod')

    m = remove_iov(m, 'ETA_4')

    assert set(m.random_variables.iov.names) == {
        'ETA_12',
        'ETA_13',
        'ETA_14',
        'ETA_15',
        'ETA_16',
        'ETA_17',
        'ETA_18',
        'ETA_19',
    }


def test_remove_iov_diagonal(create_model_for_test):
    model = create_model_for_test(
        '''$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS1
$PK
K=THETA(1)*EXP(ETA(1))+ETA(2)+ETA(3)+ETA(4)+ETA(5)+ETA(6)+ETA(7)
$ERROR
Y=F+F*EPS(1)
$THETA 0.1
$OMEGA DIAGONAL(2)
0.015
0.02
$OMEGA BLOCK(1)
0.6
$OMEGA BLOCK(1) SAME
$OMEGA 0.1
$OMEGA BLOCK(1)
0.01
$OMEGA BLOCK(1) SAME
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
'''
    )

    model = remove_iov(model)

    assert (
        '''$OMEGA DIAGONAL(2)
0.015
0.02
$OMEGA 0.1'''
        in model.model_code
    )


@pytest.mark.parametrize(
    ('distribution', 'occ', 'to_remove', 'cases', 'rest', 'abbr_ref'),
    (
        (
            'disjoint',
            'VISI',
            None,
            'IOV_1 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_1 = 0\n'
            'IOV_2 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_2 = 0\n'
            'IOV_3 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_3 = 0\n',
            (),
            '',
        ),
        (
            'joint',
            'VISI',
            None,
            'IOV_1 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_1 = 0\n'
            'IOV_2 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_2 = 0\n'
            'IOV_3 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_3 = 0\n',
            (),
            '',
        ),
        (
            'disjoint',
            'VISI',
            'ETA_IOV_1_1',
            'IOV_1 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_1 = 0\n'
            'IOV_2 = 0\n'
            'IF (VISI.EQ.3) IOV_2 = ETA_IOV_2_1\n'
            'IF (VISI.EQ.8) IOV_2 = ETA_IOV_2_2\n'
            'IOV_3 = 0\n'
            'IF (VISI.EQ.3) IOV_3 = ETA_IOV_3_1\n'
            'IF (VISI.EQ.8) IOV_3 = ETA_IOV_3_2\n',
            ('ETA_IOV_2_1', 'ETA_IOV_2_2', 'ETA_IOV_3_1', 'ETA_IOV_3_2'),
            '$ABBR REPLACE ETA_IOV_2_1=ETA(4)\n'
            '$ABBR REPLACE ETA_IOV_2_2=ETA(5)\n'
            '$ABBR REPLACE ETA_IOV_3_1=ETA(6)\n'
            '$ABBR REPLACE ETA_IOV_3_2=ETA(7)\n',
        ),
        (
            'joint',
            'VISI',
            'ETA_IOV_1_1',
            'IOV_1 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_1 = 0\n'
            'IOV_2 = 0\n'
            'IF (VISI.EQ.3) IOV_2 = ETA_IOV_2_1\n'
            'IF (VISI.EQ.8) IOV_2 = ETA_IOV_2_2\n'
            'IOV_3 = 0\n'
            'IF (VISI.EQ.3) IOV_3 = ETA_IOV_3_1\n'
            'IF (VISI.EQ.8) IOV_3 = ETA_IOV_3_2\n',
            ('ETA_IOV_2_1', 'ETA_IOV_2_2', 'ETA_IOV_3_1', 'ETA_IOV_3_2'),
            '$ABBR REPLACE ETA_IOV_2_1=ETA(4)\n'
            '$ABBR REPLACE ETA_IOV_3_1=ETA(5)\n'
            '$ABBR REPLACE ETA_IOV_2_2=ETA(6)\n'
            '$ABBR REPLACE ETA_IOV_3_2=ETA(7)\n',
        ),
        (
            'disjoint',
            'VISI',
            ['ETA_IOV_1_1', 'ETA_IOV_1_2'],
            'IOV_1 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_1 = 0\n'
            'IOV_2 = 0\n'
            'IF (VISI.EQ.3) IOV_2 = ETA_IOV_2_1\n'
            'IF (VISI.EQ.8) IOV_2 = ETA_IOV_2_2\n'
            'IOV_3 = 0\n'
            'IF (VISI.EQ.3) IOV_3 = ETA_IOV_3_1\n'
            'IF (VISI.EQ.8) IOV_3 = ETA_IOV_3_2\n',
            ('ETA_IOV_2_1', 'ETA_IOV_2_2', 'ETA_IOV_3_1', 'ETA_IOV_3_2'),
            '$ABBR REPLACE ETA_IOV_2_1=ETA(4)\n'
            '$ABBR REPLACE ETA_IOV_2_2=ETA(5)\n'
            '$ABBR REPLACE ETA_IOV_3_1=ETA(6)\n'
            '$ABBR REPLACE ETA_IOV_3_2=ETA(7)\n',
        ),
        (
            'joint',
            'VISI',
            ['ETA_IOV_1_1', 'ETA_IOV_1_2'],
            'IOV_1 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_1 = 0\n'
            'IOV_2 = 0\n'
            'IF (VISI.EQ.3) IOV_2 = ETA_IOV_2_1\n'
            'IF (VISI.EQ.8) IOV_2 = ETA_IOV_2_2\n'
            'IOV_3 = 0\n'
            'IF (VISI.EQ.3) IOV_3 = ETA_IOV_3_1\n'
            'IF (VISI.EQ.8) IOV_3 = ETA_IOV_3_2\n',
            ('ETA_IOV_2_1', 'ETA_IOV_2_2', 'ETA_IOV_3_1', 'ETA_IOV_3_2'),
            '$ABBR REPLACE ETA_IOV_2_1=ETA(4)\n'
            '$ABBR REPLACE ETA_IOV_3_1=ETA(5)\n'
            '$ABBR REPLACE ETA_IOV_2_2=ETA(6)\n'
            '$ABBR REPLACE ETA_IOV_3_2=ETA(7)\n',
        ),
        (
            'disjoint',
            'VISI',
            ['ETA_IOV_1_1', 'ETA_IOV_1_2', 'ETA_IOV_2_1'],
            'IOV_1 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_1 = 0\n'
            'IOV_2 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_2 = 0\n'
            'IOV_3 = 0\n'
            'IF (VISI.EQ.3) IOV_3 = ETA_IOV_3_1\n'
            'IF (VISI.EQ.8) IOV_3 = ETA_IOV_3_2\n',
            ('ETA_IOV_3_1', 'ETA_IOV_3_2'),
            '$ABBR REPLACE ETA_IOV_3_1=ETA(4)\n' '$ABBR REPLACE ETA_IOV_3_2=ETA(5)\n',
        ),
        (
            'joint',
            'VISI',
            ['ETA_IOV_1_1', 'ETA_IOV_1_2', 'ETA_IOV_2_1'],
            'IOV_1 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_1 = 0\n'
            'IOV_2 = 0\n'
            'IF (VISI.EQ.3.OR.VISI.EQ.8) IOV_2 = 0\n'
            'IOV_3 = 0\n'
            'IF (VISI.EQ.3) IOV_3 = ETA_IOV_3_1\n'
            'IF (VISI.EQ.8) IOV_3 = ETA_IOV_3_2\n',
            ('ETA_IOV_3_1', 'ETA_IOV_3_2'),
            '$ABBR REPLACE ETA_IOV_3_1=ETA(4)\n' '$ABBR REPLACE ETA_IOV_3_2=ETA(5)\n',
        ),
    ),
    ids=repr,
)
def test_remove_iov_with_options(
    tmp_path, load_model_for_test, testdata, distribution, occ, to_remove, cases, rest, abbr_ref
):
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox_simulated_normal.csv', tmp_path)
        model = load_model_for_test('mox2.mod')
        model = model.replace(
            datainfo=model.datainfo.replace(path=tmp_path / 'mox_simulated_normal.csv')
        )

        start_model = add_iov(model, occ=occ, distribution=distribution)

        model_with_some_iovs_removed = remove_iov(start_model, to_remove=to_remove)

        assert cases in model_with_some_iovs_removed.model_code
        assert set(model_with_some_iovs_removed.random_variables.iov.names) == set(rest)

        rec_abbr = ''.join(
            str(rec)
            for rec in model_with_some_iovs_removed.internals.control_stream.get_records(
                'ABBREVIATED'
            )
        )
        assert rec_abbr == abbr_ref


@pytest.mark.parametrize(
    'etas, etab, buf_new',
    [
        (
            ['ETA_1'],
            'ETAB1 = (EXP(ETA(1))**THETA(4) - 1)/THETA(4)',
            'CL = TVCL*EXP(ETAB1)\nV=TVV*EXP(ETA(2))',
        ),
        (
            ['ETA_1', 'ETA_2'],
            'ETAB1 = (EXP(ETA(1))**THETA(4) - 1)/THETA(4)\n'
            'ETAB2 = (EXP(ETA(2))**THETA(5) - 1)/THETA(5)',
            'CL = TVCL*EXP(ETAB1)\nV = TVV*EXP(ETAB2)',
        ),
        (
            None,
            'ETAB1 = (EXP(ETA(1))**THETA(4) - 1)/THETA(4)\n'
            'ETAB2 = (EXP(ETA(2))**THETA(5) - 1)/THETA(5)',
            'CL = TVCL*EXP(ETAB1)\nV = TVV*EXP(ETAB2)',
        ),
        (
            'ETA_1',
            'ETAB1 = (EXP(ETA(1))**THETA(4) - 1)/THETA(4)',
            'CL = TVCL*EXP(ETAB1)\nV=TVV*EXP(ETA(2))',
        ),
    ],
)
def test_transform_etas_boxcox(load_model_for_test, pheno_path, etas, etab, buf_new):
    model = load_model_for_test(pheno_path)

    model = transform_etas_boxcox(model, etas)

    rec_ref = (
        f'$PK\n'
        f'{etab}\n'
        f'IF(AMT.GT.0) BTIME=TIME\n'
        f'TAD=TIME-BTIME\n'
        f'TVCL=THETA(1)*WGT\n'
        f'TVV=THETA(2)*WGT\n'
        f'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n'
        f'{buf_new}\n'
        f'S1=V\n\n'
    )

    assert str(model.internals.control_stream.get_pred_pk_record()) == rec_ref
    assert model.parameters['lambda1'].init == 0.01


def test_transform_etas_tdist(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)

    model = transform_etas_tdist(model, ['ETA_1'])

    symbol = 'ETAT1'

    eta = 'ETA(1)'
    theta = 'THETA(4)'

    num_1 = f'{eta}**2 + 1'
    denom_1 = f'4*{theta}'

    num_2 = f'5*{eta}**4 + 16*{eta}**2 + 3'
    denom_2 = f'96*{theta}**2'

    num_3 = f'3*{eta}**6 + 17*{eta}**2 + 19*{eta}**4 - 15'
    denom_3 = f'384*{theta}**3'

    expression = (
        f'ETA(1)*(({num_1})/({denom_1}) + ({num_2})/({denom_2}) + ({num_3})/({denom_3}) + 1)'
    )

    rec_ref = (
        f'$PK\n'
        f'{symbol} = {expression}\n'
        f'IF(AMT.GT.0) BTIME=TIME\n'
        f'TAD=TIME-BTIME\n'
        f'TVCL=THETA(1)*WGT\n'
        f'TVV=THETA(2)*WGT\n'
        f'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n'
        f'CL = TVCL*EXP(ETAT1)\n'
        f'V=TVV*EXP(ETA(2))\n'
        f'S1=V\n\n'
    )

    assert str(model.internals.control_stream.get_pred_pk_record()) == rec_ref
    assert model.parameters['df1'].init == 80


@pytest.mark.parametrize(
    'etas, etad, buf_new',
    [
        (
            ['ETA_1'],
            'ETAD1 = ((ABS(ETA(1)) + 1)**THETA(4) - 1)*ABS(ETA(1))/(ETA(1)*THETA(4))',
            'CL = TVCL*EXP(ETAD1)\nV=TVV*EXP(ETA(2))',
        ),
        (
            'ETA_1',
            'ETAD1 = ((ABS(ETA(1)) + 1)**THETA(4) - 1)*ABS(ETA(1))/(ETA(1)*THETA(4))',
            'CL = TVCL*EXP(ETAD1)\nV=TVV*EXP(ETA(2))',
        ),
    ],
)
def test_transform_etas_john_draper(load_model_for_test, pheno_path, etas, etad, buf_new):
    model = load_model_for_test(pheno_path)

    model = transform_etas_john_draper(model, etas)

    rec_ref = (
        f'$PK\n'
        f'{etad}\n'
        f'IF(AMT.GT.0) BTIME=TIME\n'
        f'TAD=TIME-BTIME\n'
        f'TVCL=THETA(1)*WGT\n'
        f'TVV=THETA(2)*WGT\n'
        f'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n'
        f'{buf_new}\n'
        f'S1=V\n\n'
    )

    assert str(model.internals.control_stream.get_pred_pk_record()) == rec_ref


@pytest.mark.parametrize(
    'eta_trans,symbol,expression',
    [
        (
            EtaTransformation.boxcox(2),
            S('ETAB(2)'),
            ((sympy.exp(S('ETA(2)')) ** S('BOXCOX2') - 1) / S('BOXCOX2')),
        ),
    ],
)
def test_transform_etas_apply(eta_trans, symbol, expression):
    etas = {'eta1': S('ETA(1)'), 'eta2': S('ETA(2)'), 'etab1': S('ETAB(1)'), 'etab2': S('ETAB(2)')}

    thetas = {'theta1': 'BOXCOX1', 'theta2': 'BOXCOX2'}

    eta_trans.apply(etas, thetas)
    assert eta_trans.assignments[1].symbol == symbol
    assert eta_trans.assignments[1].expression == expression


@pytest.mark.parametrize(
    'etas, abbr_ref, omega_ref',
    [
        (
            ['ETA_CL', 'ETA_V'],
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA BLOCK(2)\n'
            '0.0309626\t; IVCL\n'
            '0.0031045\t; IIV_CL_IIV_V\n'
            '0.031128\t; IVV\n'
            '$OMEGA 0.1\n'
            '$OMEGA BLOCK(2)\n'
            '0.0309626\n'
            '0.0005 0.031128\n',
        ),
        (
            ['ETA_CL', 'ETA_S1'],
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_S1=ETA(2)\n'
            '$ABBR REPLACE ETA_V=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA BLOCK(2)\n'
            '0.0309626\t; IVCL\n'
            '0.0055644\t; IIV_CL_IIV_S1\n'
            '0.1\t; OMEGA_3_3\n'
            '$OMEGA 0.031128  ; IVV\n'
            '$OMEGA BLOCK(2)\n'
            '0.0309626\n'
            '0.0005 0.031128\n',
        ),
        (
            ['ETA_V', 'ETA_S1'],
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA 0.0309626  ; IVCL\n'
            '$OMEGA BLOCK(2)\n'
            '0.031128\t; IVV\n'
            '0.0055792\t; IIV_V_IIV_S1\n'
            '0.1\n'
            '$OMEGA BLOCK(2)\n'
            '0.0309626\n'
            '0.0005 0.031128\n',
        ),
        (
            ['ETA_CL', 'ETA_V', 'ETA_MAT'],
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_MAT=ETA(3)\n'
            '$ABBR REPLACE ETA_S1=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA BLOCK(3)\n'
            '0.0309626\t; IVCL\n'
            '0.0031045\t; IIV_CL_IIV_V\n'
            '0.031128\t; IVV\n'
            '0.0030963\t; IIV_CL_IIV_MAT\n'
            '0.0031045\t; IIV_V_IIV_MAT\n'
            '0.0309626\t; OMEGA_4_4\n'
            '$OMEGA 0.1\n'
            '$OMEGA  0.031128\n',
        ),
        (
            ['ETA_V', 'ETA_S1', 'ETA_MAT'],
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA 0.0309626  ; IVCL\n'
            '$OMEGA BLOCK(3)\n'
            '0.031128\t; IVV\n'
            '0.0055792\t; IIV_V_IIV_S1\n'
            '0.1\n'
            '0.0031045\t; IIV_V_IIV_MAT\n'
            '0.0055644\t; IIV_S1_IIV_MAT\n'
            '0.0309626\n'
            '$OMEGA  0.031128\n',
        ),
        (
            ['ETA_S1', 'ETA_MAT', 'ETA_Q'],
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA DIAGONAL(2)\n'
            '0.0309626  ; IVCL\n'
            '0.031128  ; IVV\n'
            '$OMEGA BLOCK(3)\n'
            '0.1\n'
            '0.0055644\t; IIV_S1_IIV_MAT\n'
            '0.0309626\n'
            '0.0055792\t; IIV_S1_IIV_Q\n'
            '0.0005\n'
            '0.031128\n',
        ),
        (
            None,
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA BLOCK(5)\n'
            '0.0309626\t; IVCL\n'
            '0.0031045\t; IIV_CL_IIV_V\n'
            '0.031128\t; IVV\n'
            '0.0055644\t; IIV_CL_IIV_S1\n'
            '0.0055792\t; IIV_V_IIV_S1\n'
            '0.1\n'
            '0.0030963\t; IIV_CL_IIV_MAT\n'
            '0.0031045\t; IIV_V_IIV_MAT\n'
            '0.0055644\t; IIV_S1_IIV_MAT\n'
            '0.0309626\n'
            '0.0031045\t; IIV_CL_IIV_Q\n'
            '0.0031128\t; IIV_V_IIV_Q\n'
            '0.0055792\t; IIV_S1_IIV_Q\n'
            '0.0005\n'
            '0.031128\n',
        ),
    ],
)
def test_create_joint_distribution_plain(load_model_for_test, testdata, etas, abbr_ref, omega_ref):
    model_start = load_model_for_test(testdata / 'nonmem/pheno_block.mod')

    model = create_joint_distribution(model_start, etas, individual_estimates=None)

    rec_abbr = ''.join(
        str(rec) for rec in model.internals.control_stream.get_records('ABBREVIATED')
    )
    assert rec_abbr == abbr_ref

    rec_pk = str(model.internals.control_stream.get_pred_pk_record())
    pk_ref = str(model_start.internals.control_stream.get_pred_pk_record())
    assert rec_pk == pk_ref

    rec_omega = ''.join(str(rec) for rec in model.internals.control_stream.get_records('OMEGA'))
    assert rec_omega == omega_ref


@pytest.mark.parametrize(
    'etas, abbr_ref, omega_ref',
    [
        (
            (['ETA_CL', 'ETA_V'], ['ETA_CL', 'ETA_S1']),
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_S1=ETA(2)\n'
            '$ABBR REPLACE ETA_V=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA BLOCK(2)\n'
            '0.0309626\t; IVCL\n'
            '0.0055644\t; IIV_CL_IIV_S1\n'
            '0.1\t; OMEGA_3_3\n'
            '$OMEGA  0.031128 ; IVV\n'
            '$OMEGA BLOCK(2)\n'
            '0.0309626\n'
            '0.0005 0.031128\n',
        ),
        (
            (None, ['ETA_CL', 'ETA_V']),
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA BLOCK(2)\n'
            '0.0309626\t; IVCL\n'
            '0.0031045\t; IIV_CL_IIV_V\n'
            '0.031128\t; IVV\n'
            '$OMEGA BLOCK(3)\n'
            '0.1\n'
            '0.0055644\t; IIV_S1_IIV_MAT\n'
            '0.0309626\n'
            '0.0055792\t; IIV_S1_IIV_Q\n'
            '0.0005\n'
            '0.031128\n',
        ),
        (
            (['ETA_CL', 'ETA_V'], None),
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA BLOCK(5)\n'
            '0.0309626\t; IVCL\n'
            '0.0031045\t; IIV_CL_IIV_V\n'
            '0.031128\t; IVV\n'
            '0.0055644\t; IIV_CL_IIV_S1\n'
            '0.0055792\t; IIV_V_IIV_S1\n'
            '0.1\n'
            '0.0030963\t; IIV_CL_IIV_MAT\n'
            '0.0031045\t; IIV_V_IIV_MAT\n'
            '0.0055644\t; IIV_S1_IIV_MAT\n'
            '0.0309626\n'
            '0.0031045\t; IIV_CL_IIV_Q\n'
            '0.0031128\t; IIV_V_IIV_Q\n'
            '0.0055792\t; IIV_S1_IIV_Q\n'
            '0.0005\n'
            '0.031128\n',
        ),
    ],
)
def test_create_joint_distribution_nested(load_model_for_test, testdata, etas, abbr_ref, omega_ref):
    model_start = load_model_for_test(testdata / 'nonmem/pheno_block.mod')

    model = create_joint_distribution(model_start, etas[0], individual_estimates=None)
    model = create_joint_distribution(model, etas[1], individual_estimates=None)

    rec_abbr = ''.join(
        str(rec) for rec in model.internals.control_stream.get_records('ABBREVIATED')
    )
    assert rec_abbr == abbr_ref

    rec_pk = str(model.internals.control_stream.get_pred_pk_record())
    pk_ref = str(model_start.internals.control_stream.get_pred_pk_record())
    assert rec_pk == pk_ref

    rec_omega = ''.join(str(rec) for rec in model.internals.control_stream.get_records('OMEGA'))

    assert rec_omega == omega_ref

    cov_params = {p.symbol for p in model.parameters if p.symbol.name.startswith('IIV')}
    assert all(p in model.random_variables.free_symbols for p in cov_params)
    assert not re.search(r'\$THETA\s+0\.\d+\s+;\s+IIV_', model.model_code)


@pytest.mark.parametrize(
    'etas, abbr_ref, omega_ref',
    [
        (
            ['ETA_CL'],
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA  0.0309626 ; IVCL\n'
            '$OMEGA BLOCK(4)\n'
            '0.031128\t; IVV\n'
            '0.0055792\t; IIV_V_IIV_S1\n'
            '0.1\n'
            '0.0031045\t; IIV_V_IIV_MAT\n'
            '0.0055644\t; IIV_S1_IIV_MAT\n'
            '0.0309626\n'
            '0.0031128\t; IIV_V_IIV_Q\n'
            '0.0055792\t; IIV_S1_IIV_Q\n'
            '0.0005\n'
            '0.031128\n',
        ),
        (
            ['ETA_CL', 'ETA_V'],
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA  0.0309626 ; IVCL\n'
            '$OMEGA  0.031128 ; IVV\n'
            '$OMEGA BLOCK(3)\n'
            '0.1\n'
            '0.0055644\t; IIV_S1_IIV_MAT\n'
            '0.0309626\n'
            '0.0055792\t; IIV_S1_IIV_Q\n'
            '0.0005\n'
            '0.031128\n',
        ),
        (
            ['ETA_CL', 'ETA_S1'],
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_S1=ETA(2)\n'
            '$ABBR REPLACE ETA_V=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA  0.0309626 ; IVCL\n'
            '$OMEGA  0.1 ; OMEGA_3_3\n'
            '$OMEGA BLOCK(3)\n'
            '0.031128\t; IVV\n'
            '0.0031045\t; IIV_V_IIV_MAT\n'
            '0.0309626\n'
            '0.0031128\t; IIV_V_IIV_Q\n'
            '0.0005\n'
            '0.031128\n',
        ),
        (
            None,
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA  0.0309626 ; IVCL\n'
            '$OMEGA  0.031128 ; IVV\n'
            '$OMEGA  0.1\n'
            '$OMEGA  0.0309626\n'
            '$OMEGA  0.031128\n',
        ),
        (
            'ETA_CL',
            '$ABBR REPLACE ETA_CL=ETA(1)\n'
            '$ABBR REPLACE ETA_V=ETA(2)\n'
            '$ABBR REPLACE ETA_S1=ETA(3)\n'
            '$ABBR REPLACE ETA_MAT=ETA(4)\n'
            '$ABBR REPLACE ETA_Q=ETA(5)\n',
            '$OMEGA  0.0309626 ; IVCL\n'
            '$OMEGA BLOCK(4)\n'
            '0.031128\t; IVV\n'
            '0.0055792\t; IIV_V_IIV_S1\n'
            '0.1\n'
            '0.0031045\t; IIV_V_IIV_MAT\n'
            '0.0055644\t; IIV_S1_IIV_MAT\n'
            '0.0309626\n'
            '0.0031128\t; IIV_V_IIV_Q\n'
            '0.0055792\t; IIV_S1_IIV_Q\n'
            '0.0005\n'
            '0.031128\n',
        ),
    ],
)
def test_split_joint_distribution(load_model_for_test, testdata, etas, abbr_ref, omega_ref):
    model_start = load_model_for_test(testdata / 'nonmem/pheno_block.mod')
    model = create_joint_distribution(model_start)

    model = split_joint_distribution(model, etas)

    rec_abbr = ''.join(
        str(rec) for rec in model.internals.control_stream.get_records('ABBREVIATED')
    )
    assert rec_abbr == abbr_ref

    rec_pk = str(model.internals.control_stream.get_pred_pk_record())
    pk_ref = str(model_start.internals.control_stream.get_pred_pk_record())
    assert rec_pk == pk_ref

    rec_omega = ''.join(str(rec) for rec in model.internals.control_stream.get_records('OMEGA'))

    assert rec_omega == omega_ref


@pytest.mark.parametrize(
    'rvs, exception_msg',
    [
        (['ETA_3', 'NON_EXISTENT_RV'], r'.*non-existing.*'),
        (['ETA_3', 'ETA_6'], r'.*ETA_6.*'),
        (['ETA_1'], 'At least two random variables are needed'),
    ],
)
def test_create_joint_distribution_incorrect_params(
    load_model_for_test, testdata, rvs, exception_msg
):
    model = load_model_for_test(
        testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'multEST' / 'noSIM' / 'withBayes.mod'
    )

    with pytest.raises(Exception, match=exception_msg):
        create_joint_distribution(model, rvs, individual_estimates=None)


def test_create_joint_distribution_choose_param_init(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    res = read_modelfit_results(pheno_path)
    params = (model.parameters['IVCL'], model.parameters['IVV'])
    rvs = model.random_variables.etas
    init = _choose_cov_param_init(model, res.individual_estimates, rvs, *params)
    assert init == 0.0118179

    model = load_model_for_test(pheno_path)
    init = _choose_cov_param_init(model, None, rvs, *params)
    assert init == 0.0031045

    model = load_model_for_test(pheno_path)
    rv_new = NormalDistribution.create('ETA_3', 'IIV', 0, S('OMEGA_3_3'))
    rvs += rv_new
    ie = res.individual_estimates.copy()
    ie['ETA_3'] = ie['ETA_1']
    init = _choose_cov_param_init(model, res.individual_estimates, rvs, *params)
    assert init == 0.0118179

    # If one eta doesn't have individual estimates
    model = load_model_for_test(pheno_path)
    model = add_iiv(model, 'S1', 'add')
    params = (model.parameters['IVCL'], model.parameters['IIV_S1'])
    rvs = model.random_variables[('ETA_1', 'ETA_S1')]
    init = _choose_cov_param_init(model, res.individual_estimates, rvs, *params)
    assert init == 0.0052789

    # If the standard deviation in individual estimates of one eta is 0
    model = load_model_for_test(pheno_path)
    ie = res.individual_estimates.copy()
    ie['ETA_1'] = 0
    params = (model.parameters['IVCL'], model.parameters['IVV'])
    rvs = model.random_variables[('ETA_1', 'ETA_2')]
    with pytest.warns(UserWarning, match='Correlation of individual estimates'):
        init = _choose_cov_param_init(model, ie, rvs, *params)
        assert init == 0.0031045


def test_create_joint_distribution_choose_param_init_fo(create_model_for_test):
    model = create_model_for_test(
        '''$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(1))
S1=V*EXP(ETA(2))

$ERROR
Y=F+F*EPS(1)

$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.1  ; IVCL
$OMEGA 0.1  ; IVV
$SIGMA 0.013241

$ESTIMATION METHOD=0
'''
    )
    params = (model.parameters['IVCL'], model.parameters['IVV'])
    rvs = model.random_variables.etas
    init = _choose_cov_param_init(model, None, rvs, *params)

    assert init == 0.01


def test_create_joint_distribution_names(create_model_for_test):
    model = create_model_for_test(
        '''$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(1))
S1=V*EXP(ETA(2))

$ERROR
Y=F+F*EPS(1)

$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241

$ESTIMATION METHOD=1 INTERACTION
'''
    )
    model = create_joint_distribution(
        model,
        model.random_variables.names,
        individual_estimates=model.modelfit_results.individual_estimates
        if model.modelfit_results is not None
        else None,
    )
    assert 'IIV_CL_V_IIV_S1' in model.parameters.names


@pytest.mark.parametrize(
    'etas_file, force, file_exists',
    [('', False, False), ('', True, True), ('$ETAS FILE=run1.phi', False, True)],
)
def test_update_initial_individual_estimates(
    load_model_for_test, testdata, etas_file, force, file_exists, tmp_path
):
    shutil.copy(testdata / 'nonmem/pheno.mod', tmp_path / 'run1.mod')
    shutil.copy(testdata / 'nonmem/pheno.phi', tmp_path / 'run1.phi')
    shutil.copy(testdata / 'nonmem/pheno.ext', tmp_path / 'run1.ext')
    shutil.copy(testdata / 'nonmem/pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        with open('run1.mod', 'a') as f:
            f.write(etas_file)

        model = load_model_for_test('run1.mod')
        res = read_modelfit_results('run1.mod')
        model = update_initial_individual_estimates(model, res.individual_estimates, force=force)
        model = model.write_files()

        assert ('$ETAS FILE=run1_input.phi' in model.model_code) is file_exists
        assert (os.path.isfile('run1_input.phi')) is file_exists


def test_nested_iiv_transformations(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    res = read_modelfit_results(pheno_path)

    model = create_joint_distribution(model, individual_estimates=res.individual_estimates)

    assert 'IIV_CL_IIV_V' in model.model_code

    model = load_model_for_test(pheno_path)

    model = remove_iiv(model, 'CL')

    assert '0.031128' in model.model_code
    assert '0.0309626' not in model.model_code

    model = load_model_for_test(pheno_path)

    model = remove_iiv(model, 'V')

    assert '0.0309626' in model.model_code
    assert '0.031128' not in model.model_code
